/*
 * $Id: common.c,v 1.6 1998/06/03 21:38:55 masaki Exp $
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
extern char *addr2ascii(int, const void *, int, char *);

#include <mrt.h>
#include <New.h>
#include <linked_list.h>
#include <interface.h>
#ifdef HAVE_IPV6
#include <api6.h>
#endif /* HAVE_IPV6 */

interface_master_t *INTERFACE_MASTER;
char *print_iflags (char *tmpx, int len, u_long flags);


interface_t *
new_interface (char *name, u_long flags, int mtu, int index)
{
    interface_t *interface;
    char tmpx[MAXLINE];

    assert (INTERFACE_MASTER);

    if ((interface = find_interface_byname (name))) {
	return (interface);
    }
    if (index >= MAX_INTERFACES) {
        trace (TR_FATAL, INTERFACE_MASTER->trace,
	   "Too many interfaces %d (should be < %d)\n", 
	   INTERFACE_MASTER->number, MAX_INTERFACES);
	return (NULL);
    }

    if (index <= 0)
	index = INTERFACE_MASTER->number + 1;

    interface = New (interface_t);
    memset (interface, 0, sizeof (*interface));
    LL_Add (INTERFACE_MASTER->ll_interfaces, interface);

    interface->ll_addr = LL_Create (0);
    interface->flags = flags;
    interface->mtu = mtu;
    memset (interface->dlist_in, 0, sizeof (interface->dlist_in));
    memset (interface->dlist_out, 0, sizeof (interface->dlist_out));
    interface->metric_in = 1;
    interface->metric_out = 0;
    strcpy (interface->name, name);

    /* set bit for easy identification */
    assert (INTERFACE_MASTER->number < sizeof (interface->bit) * 8);
    interface->bit = 1 << INTERFACE_MASTER->number;
    INTERFACE_MASTER->number++;
    interface->index = index;
    INTERFACE_MASTER->index2if[index] = interface;

    if (INTERFACE_MASTER->max_mtu < mtu)
	INTERFACE_MASTER->max_mtu = mtu;

    trace (TR_INFO, INTERFACE_MASTER->trace,
	   "Interface %s flags %s mtu %d index %d\n",
	   name, print_iflags (tmpx, sizeof (tmpx), flags), mtu, index);

    return (interface);
}


ll_addr_t *
New_Addr (int family, void *addr, int bitlen, void *broadcast)
{

    ll_addr_t *if_addr;

    if_addr = New (ll_addr_t);
    assert (addr);
    if_addr->prefix = New_Prefix (family, addr, bitlen);
    if (broadcast)
	if_addr->broadcast = New_Prefix (family, broadcast, bitlen);
    else
	if_addr->broadcast = NULL;

    return (if_addr);
}


interface_t *
add_addr_to_interface (interface_t * interface, int family,
		       void *addr, int len, void *broadcast)
{

    char tmpx[MAXLINE];
    ll_addr_t *if_addr;

    assert (interface);
    assert (INTERFACE_MASTER);

#ifdef HAVE_IPV6
    assert (family == AF_INET || family == AF_INET6);
#else
    assert (family == AF_INET);
#endif /* HAVE_IPV6 */

#ifdef HAVE_IPV6
    /* we don't accept these ones */
    if (family == AF_INET6) {
	char tmp6[INET6_ADDRSTRLEN];
	char tmp7[INET6_ADDRSTRLEN];

	if (IN6_IS_ADDR_V4MAPPED ((struct in6_addr *) addr) ||
	    IN6_IS_ADDR_V4COMPAT ((struct in6_addr *) addr)) {
	    trace (TR_WARN, INTERFACE_MASTER->trace,
	           "Interface %s inet6 %s/%d skipped\n", interface->name,
	           inet_ntop (family, addr, tmp6, sizeof tmp6), len);
	    return (NULL);
	}
	if (broadcast && (
	    IN6_IS_ADDR_V4MAPPED ((struct in6_addr *) broadcast) ||
	    IN6_IS_ADDR_V4COMPAT ((struct in6_addr *) broadcast))) {
	    trace (TR_WARN, INTERFACE_MASTER->trace,
	           "Interface %s inet6 %s/%d %s %s skipped\n", interface->name,
	           inet_ntop (family, addr, tmp6, sizeof tmp6), len,
		     BIT_TEST (interface->flags, IFF_POINTOPOINT) ?
		     "dest" : "broadcast",
		     inet_ntop (family, broadcast, tmp7, sizeof tmp7));
	    return (NULL);
	}
    }
#endif

    if_addr = New_Addr (family, addr, len, broadcast);
    LL_Add (interface->ll_addr, if_addr);

    if (family == AF_INET && interface->primary == NULL) {
	interface->primary = if_addr;
    }
#ifdef HAVE_IPV6
    else if (family == AF_INET6) {
	if (IN6_IS_ADDR_V4MAPPED ((struct in6_addr *) addr) &&
	    IN6_IS_ADDR_V4COMPAT ((struct in6_addr *) addr)) {
	    return (NULL);
	}

	if (interface->primary6 == NULL || (!IN6_IS_ADDR_UC_GLOBAL (
			    prefix_toaddr6 (interface->primary6->prefix)) &&
		     IN6_IS_ADDR_UC_GLOBAL ((struct in6_addr *) addr))) {
	    interface->primary6 = if_addr;
	}
	if (interface->link_local == NULL &&
	    IN6_IS_ADDR_LINKLOCAL ((struct in6_addr *) addr)) {
	    interface->link_local = if_addr;
	}
    }
#endif /* HAVE_IPV6 */

#ifdef mrt_compat
    if (family == AF_INET && interface->prefix == NULL) {
	interface->prefix = New_Prefix (family, addr, len);
	if (broadcast)
	    memcpy (&interface->broadcast, broadcast, 4);
    }
#ifdef HAVE_IPV6
    else if (family == AF_INET6 && interface->prefix6 == NULL) {
	interface->prefix6 = New_Prefix (family, addr, len);
	if (broadcast)
	    memcpy (&interface->dest6, broadcast, 16);
    }
#endif /* HAVE_IPV6 */
#endif

#ifdef HAVE_IPV6
    /* the first non-loopback ipv6 capable interface becomes our v6 default */
    if (family == AF_INET6 &&
	strncmp (interface->name, "sit", 3) != 0 &&	/* XXX */
	INTERFACE_MASTER->default_interface6 == NULL &&
	!BIT_TEST (interface->flags, IFF_LOOPBACK) &&
	ipv6_link_local_addr ((struct in6_addr *) addr)) {
	INTERFACE_MASTER->default_interface6 = interface;
	trace (TR_INFO, INTERFACE_MASTER->trace,
	       "DEFAULT V6 Interface %s\n", interface->name);
    }
#endif /* HAVE_IPV6 */
    /* the first non-loopback interface becomes our v4 default */
    if (family == AF_INET &&
	INTERFACE_MASTER->default_interface == NULL &&
	!BIT_TEST (interface->flags, IFF_TUNNEL) &&
	!BIT_TEST (interface->flags, IFF_LOOPBACK)) {
	INTERFACE_MASTER->default_interface = interface;
	trace (TR_INFO, INTERFACE_MASTER->trace,
	       "DEFAULT V4 Interface %s\n", interface->name);
    }

    /* choose the largest ip number */
    else if (family == AF_INET &&
	INTERFACE_MASTER->default_interface &&
	!BIT_TEST (interface->flags, IFF_LOOPBACK) &&
	ntohl (prefix_tolong (
		INTERFACE_MASTER->default_interface->primary->prefix)) 
	    < ntohl (prefix_tolong (interface->primary->prefix))) {
	INTERFACE_MASTER->default_interface = interface;
	trace (TR_INFO, INTERFACE_MASTER->trace,
	       "DEFAULT V4 Interface updated with %s\n", interface->name);
    }

    tmpx[0] = '\0';
    if (family == AF_INET) {
	char tmp4[16];
	if (broadcast)
	    sprintf (tmpx, " %s %s", 
		     BIT_TEST (interface->flags, IFF_POINTOPOINT) ?
		     "dest" : "broadcast",
		     inet_ntop (family, broadcast, tmp4, sizeof tmp4));
	trace (TR_INFO, INTERFACE_MASTER->trace,
	       "Interface %s inet %s/%d%s\n", interface->name,
	       inet_ntop (family, addr, tmp4, sizeof tmp4),
	       len, tmpx);
    }
#ifdef HAVE_IPV6
    else if (family == AF_INET6) {
	char tmp6[INET6_ADDRSTRLEN];
	if (broadcast)
	    sprintf (tmpx, " %s %s", 
		     BIT_TEST (interface->flags, IFF_POINTOPOINT) ?
		     "dest" : "broadcast",
		     inet_ntop (family, broadcast, tmp6, sizeof tmp6));
	trace (TR_INFO, INTERFACE_MASTER->trace,
	       "Interface %s inet6 %s/%d%s\n", interface->name,
	       inet_ntop (family, addr, tmp6, sizeof tmp6),
	       len, tmpx);
    }
#endif /* HAVE_IPV6 */

    return (interface);
}


int
init_interfaces (trace_t * ltrace)
{
    assert (INTERFACE_MASTER == NULL);
    INTERFACE_MASTER = New (interface_master_t);
    INTERFACE_MASTER->ll_interfaces = LL_Create (0);
    INTERFACE_MASTER->trace = trace_copy (ltrace);
    set_trace (INTERFACE_MASTER->trace, TRACE_PREPEND_STRING, "IF", 0);
    INTERFACE_MASTER->number = 0;
    INTERFACE_MASTER->max_mtu = 576;
    INTERFACE_MASTER->default_interface = NULL;
    if ((INTERFACE_MASTER->sockfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
        trace (TR_FATAL, INTERFACE_MASTER->trace,
               "socket for AF_INET (%s)\n", strerror (errno));
    }
#ifdef HAVE_IPV6
    if ((INTERFACE_MASTER->sockfd6 = socket (AF_INET6, SOCK_DGRAM, 0)) < 0) {
        trace (TR_WARN, INTERFACE_MASTER->trace,
               "socket for AF_INET6 (%s)\n", strerror (errno));
    }
#endif /* HAVE_IPV6 */
    read_interfaces ();

    if (INTERFACE_MASTER->default_interface == NULL) {
        trace (TR_FATAL, INTERFACE_MASTER->trace, 
	       "Interface initialization Failed, Aborting\n");
	return (-1);
    }
    return (1);
}


/* 
 * Given a prefix, find an interface that prefix is on
 */
interface_t *
find_interface (prefix_t * prefix)
{

    return (find_interface_flags (prefix, 0));
}


/* 
 * Given a prefix, find an interface with the specified flags 
 * that prefix is on if flags == 0, then see everything
 */
interface_t *
find_interface_flags (prefix_t * prefix, u_long flags)
{
    interface_t *interface;
    ll_addr_t *ll_addr;
    u_char *addr, *dest;
    int llocal = 0;

    /* bgpsim doesn't initialize ifs */
    if (INTERFACE_MASTER == NULL)
	return (NULL);

    addr = prefix_tochar (prefix);
#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6 &&
            ipv6_link_local_addr ((struct in6_addr *) addr)) {
	llocal++;
    }
#endif /* HAVE_IPV6 */
    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {

	if (flags && !BIT_TEST (interface->flags, flags))
	    continue;

	if (interface->ll_addr) {
	    LL_Iterate (interface->ll_addr, ll_addr) {
		if (ll_addr->prefix == NULL)
		    continue;
		if (ll_addr->prefix->family != prefix->family)
		    continue;
		if (!llocal && ll_addr->prefix->bitlen > prefix->bitlen)
		    continue;

		dest = prefix_tochar (ll_addr->prefix);
		if (BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
		    ll_addr->broadcast) {
		    /* Linux's sit is p-to-p but doesn't have dest addr */
		    dest = prefix_tochar (ll_addr->broadcast);
		}

#if 0
		printf ("name = %s ", interface->name);
		if (ll_addr->prefix->family == AF_INET6) {
		    printf ("addr = %s ", inet_ntop (AF_INET6, addr, tmp6, sizeof tmp6));

		    printf ("dest = %s ", inet_ntop (AF_INET6, dest, tmp6, sizeof tmp6));
		}
		else {
		    printf ("addr = %s ", inet_ntop (AF_INET, addr, tmp6, sizeof tmp6));
		    printf ("dest = %s ", inet_ntop (AF_INET, dest, tmp6, sizeof tmp6));
		}
		printf ("bitlen = %d\n", ll_addr->prefix->bitlen);
#endif
		if (addr && dest && comp_with_mask (addr, dest,
				  (llocal) ? 128 : ll_addr->prefix->bitlen))
		    return (interface);

	    }
	}
    }

    return (NULL);
}


/*
 * Given a prefix, find interfaces on that prefix's network
 *   matches exactly or includes
 */
LINKED_LIST *
find_network (prefix_t * prefix)
{
    interface_t *interface;
    ll_addr_t *ll_addr;
    u_char *addr, *dest;
    LINKED_LIST *ll = NULL;

    if (INTERFACE_MASTER == NULL) return (NULL);

    addr = prefix_tochar (prefix);

    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
#ifdef notdef
	if (BIT_TEST (interface->flags, IFF_LOOPBACK))
	    continue;
#endif

	if (interface->ll_addr) {
	    LL_Iterate (interface->ll_addr, ll_addr) {
		if (ll_addr->prefix == NULL)
		    continue;
		if (ll_addr->prefix->family != prefix->family)
		    continue;
		if (ll_addr->prefix->bitlen < prefix->bitlen)
		    continue;

		dest = prefix_tochar (ll_addr->prefix);
		if (BIT_TEST (interface->flags, IFF_POINTOPOINT)) {
		    if (ll_addr->broadcast == NULL)
			continue;
		    dest = prefix_tochar (ll_addr->broadcast);
		}

#if 0
		printf ("name = %s ", interface->name);
		if (ll_addr->prefix->family == AF_INET6) {
		    printf ("addr = %s ", inet_ntop (AF_INET6, addr, NULL, 0));
		    printf ("dest = %s ", inet_ntop (AF_INET6, dest, NULL, 0));
		}
		else {
		    printf ("addr = %s ", inet_ntop (AF_INET, addr, NULL, 0));
		    printf ("dest = %s ", inet_ntop (AF_INET, dest, NULL, 0));
		}
		printf ("bitlen = %d\n", prefix->bitlen);
#endif
		if (addr && dest &&
		    comp_with_mask (addr, dest, prefix->bitlen)) {
		    if (ll == NULL)
			ll = LL_Create (0);
		    LL_Add (ll, interface);
		    break;
		}
	    }
	}
    }

    return (ll);
}


/* 
 * Given a name, find an interface that has the name
 */
interface_t *
find_interface_byname (char *name)
{
    interface_t *interface;

    if (INTERFACE_MASTER == NULL) return (NULL);

    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	if (strcasecmp (name, interface->name) == 0)
	    return (interface);
    }
    return (NULL);
}


/* 
 * if it belongs to a local interface, its pointer returned
 * this is used to detect when we have received a broadcast packet from 
 * ourselves in rip, and ripng
 */
interface_t *
find_interface_local (prefix_t * prefix)
{
    interface_t *interface;
    ll_addr_t *ll_addr;
    int family = prefix->family;
    u_char *addr = prefix_tochar (prefix);

    if (INTERFACE_MASTER == NULL) return (NULL);

#ifndef HAVE_IPV6
    assert (prefix->family == AF_INET);
#else
    assert (prefix->family == AF_INET || prefix->family == AF_INET6);
#endif /* HAVE_IPV6 */

    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	if (interface->ll_addr) {
	    LL_Iterate (interface->ll_addr, ll_addr) {
		if (ll_addr->prefix == NULL)
		    continue;
		if (ll_addr->prefix->family != family)
		    continue;

		/*
		 * The local side addresses are checked even in case of p-to-p
		 */
		addr = prefix_tochar (ll_addr->prefix);
		if (memcmp (prefix_tochar (prefix), addr, (family == AF_INET) ? 4 : 16) == 0)
		    return (interface);
	    }
	}
    }

    return (NULL);
}


int
is_prefix_on (prefix_t *prefix, interface_t *interface)
{
    ll_addr_t *ll_addr;
    u_char *addr, *dest;
    int bitlen;

    assert (prefix);
    assert (interface);

    if (interface->ll_addr == NULL)
	return (0);

    addr = prefix_tochar (prefix);

    LL_Iterate (interface->ll_addr, ll_addr) {
	if (ll_addr->prefix == NULL)
	    continue;
	if (ll_addr->prefix->family != prefix->family)
	    continue;
/*
	if (ll_addr->prefix->bitlen > prefix->bitlen)
	    continue;
*/
	dest = prefix_tochar (ll_addr->prefix);
	bitlen = ll_addr->prefix->bitlen;

	if (BIT_TEST (interface->flags, IFF_POINTOPOINT)) {
	    /* there is no destionation addr for sit on Linux */
	    if (ll_addr->broadcast == NULL)
		continue;
	    dest = prefix_tochar (ll_addr->broadcast);
	    bitlen = 32;
#ifdef HAVE_IPV6
	    if (ll_addr->broadcast->family == AF_INET6)
	        bitlen = 128;
#endif /* HAVE_IPV6 */
	}

	if (comp_with_mask (addr, dest, bitlen))
	    return (1);
    }
    return (0);
}


/* 
 * compatible purpose
 */
interface_t *
local_interface (int family, void *cp)
{
    prefix_t ptmp, *prefix = &ptmp;

    if (INTERFACE_MASTER == NULL) return (NULL);

    if ((prefix->family = family) == AF_INET) {
	prefix->bitlen = 32;
	memcpy (&prefix->add.sin, cp, 4);
    }
#ifdef HAVE_IPV6
    else if (family == AF_INET6) {
	prefix->bitlen = 128;
	memcpy (&prefix->add.sin6, cp, 16);
    }
#endif /* HAVE_IPV6 */

    return (find_interface_local (prefix));
}


/* show_interfaces
 * Dump various interface stats/info to a socket
 */
int
show_interfaces (uii_connection_t * uii)
{
    interface_t *interface;
    ll_addr_t *ll_addr;
    char tmpx[MAXLINE];

    if (INTERFACE_MASTER == NULL) return (0);

    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	uii_send_data (uii, "Interface %s %s mtu %d weight %d\r\n",
		       interface->name,
		       print_iflags (tmpx, MAXLINE, interface->flags),
		       interface->mtu, interface->default_pref);

#ifdef notdef
	if (interface->prefix != NULL) {
	    uii_send_data (uii, "  inet4 %s\r\n", prefix_toa (interface->prefix));
	}
#ifdef HAVE_IPV6
	if (interface->prefix6 != NULL) {
	    uii_send_data (uii, "  inet6 %s\r\n", prefix_toa (interface->prefix6));
	}
#endif /* HAVE_IPV6 */
#endif
	LL_Iterate (interface->ll_addr, ll_addr) {
	    assert (ll_addr->prefix);

	    tmpx[0] = '\0';
	    if (ll_addr->broadcast)
		sprintf (tmpx, " %s %s", BIT_TEST (interface->flags, IFF_POINTOPOINT) ?
		      "-->" : "broadcast", prefix_toa (ll_addr->broadcast));

	    uii_send_data (uii, "  %s %s/%d%s\r\n",
			   (ll_addr->prefix->family == AF_INET) ? "inet" :
#ifdef HAVE_IPV6
			   (ll_addr->prefix->family == AF_INET6) ? "inet6" :
#endif /* HAVE_IPV6 */
			   "???",
		prefix_toa (ll_addr->prefix), ll_addr->prefix->bitlen, tmpx);
	}
	uii_send_data (uii, "\r\n");
    }
    return (1);
}


typedef struct {
    u_int bit;
    char *name;
} bits;
static bits iflags[] =
{
    {IFF_UP, "UP"},
    {IFF_BROADCAST, "BROADCAST"},
    {IFF_DEBUG, "DEBUG"},
    {IFF_LOOPBACK, "LOOPBACK"},
    {IFF_POINTOPOINT, "POINTOPOINT"},
#ifdef  IFF_NOTRAILERS
    {IFF_NOTRAILERS, "NOTRAILERS"},
#endif				/* IFF_NOTRAILERS */
    {IFF_RUNNING, "RUNNING"},
#ifdef  IFF_NOARP
    {IFF_NOARP, "NOARP"},
#endif				/* IFF_NOARP */
    {IFF_PROMISC, "PROMISC"},
    {IFF_ALLMULTI, "ALLMULTI"},
#ifdef  IFF_OACTIVE
    {IFF_OACTIVE, "OACTIVE"},
#endif				/* IFF_OACTIVE */
#ifdef  IFF_SIMPLEX
    {IFF_SIMPLEX, "SIMPLEX"},
#endif				/* IFF_SIMPLEX */
#ifdef  IFF_LINK0
    {IFF_LINK0, "LINK0"},
    {IFF_LINK1, "LINK1"},
    {IFF_LINK2, "LINK2"},
#endif				/* IFF_LINK0 */
#ifdef  IFF_MULTICAST
    {IFF_MULTICAST, "MULTICAST"},
#endif				/* IFF_MULTICAST */
#ifdef  IFF_USERDEF
    {IFF_USERDEF, "USERDEF"},
#endif				/* IFF_USERDEF */
#ifdef  IFF_TUNNEL
    {IFF_TUNNEL, "TUNNEL"},
#endif				/* IFF_TUNNEL */
    {0, NULL}
};


char *
print_iflags (char *tmpx, int len, u_long flags)
{
    int i, l, n = 0;

    assert (len >= 3);
    tmpx[n++] = '<';
    for (i = 0; iflags[i].bit; i++) {
	if (!BIT_TEST (flags, iflags[i].bit))
	    continue;
	if (n > 1)
	    tmpx[n++] = ',';
	l = strlen (iflags[i].name);
	if (n + l + 1 > len - 1)
	    break;
	strcpy (tmpx + n, iflags[i].name);
	n += l;
    }
    tmpx[n++] = '>';
    tmpx[n] = '\0';
    return (tmpx);
}


int 
mask2len (void *vmask, int bytes)
{
    int i, j, bitlen = 0;
    u_char *mask = vmask;

    if (bytes == 0 || mask == NULL)
	return (bitlen);
    for (i = 0; i < bytes; i++, bitlen += 8) {
	if (mask[i] != 0xff)
	    break;
    }
    if (i != bytes && mask[i]) {
	for (j = 7; j > 0; j--, bitlen++) {
	    if ((mask[i] & (1 << j)) == 0)
		break;
	}
    }
    /* this doesn't check if the rest of bits are all 0 or not */
    /* test in another place if contiguous or not */

#if 0
    if (bitlen == 0)
	bitlen = bytes * 8;
#endif
    return (bitlen);
}


u_char *
len2mask (int bitlen, void *vmask, int bytes)
{
    int i;
    u_char *mask = vmask;

    assert (bytes * 8 >= bitlen);
    if (bitlen == 0) {
	memset (mask, 0, bytes);
	return (mask);
    }
    if (mask == NULL || bytes == 0)
	return (NULL);
    for (i = 0; i < (bitlen / 8); i++)
	mask[i] = 0xff;
    if (bitlen < (bytes * 8))
	mask[i++] = 0xff << (8 - (bitlen & 7));
    for (; i < bytes; i++)
	mask[i] = 0;
    return (mask);
}


interface_t *
find_interface_byindex (int index)
{
    if (INTERFACE_MASTER == NULL) return (NULL);

#ifdef notdef
    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	if (interface->index == index)
	    return (interface);
    }
    return (NULL);
#else
    assert (index < MAX_INTERFACES);
    return (INTERFACE_MASTER->index2if[index]);
#endif
}


int 
add_kernel_route (int family, void *dest, void *nhop, int masklen, int index)
{
    prefix_t *prefix, *nexthop;
    int len = 32;
    interface_t *interface = NULL;

#ifdef HAVE_IPV6
    assert (family == AF_INET || family == AF_INET6);
#else
    assert (family == AF_INET);
#endif /* HAVE_IPV6 */

    if (family == AF_INET) {
	if (*(u_long *) nhop == INADDR_ANY)
	    return (0);
    }
#ifdef HAVE_IPV6
    else {
        struct in6_addr zero6;

	/* XXX I haven't coded the same thing in IPv4 */
	if (index > 0) {
	    interface = find_interface_byindex (index);
	}

	memset (&zero6, 0, sizeof (zero6));
	/* linux way of direct connected route */
	if (memcmp (nhop, &zero6, 16) == 0)
	    return (0);
	/*
	 * I'm not sure but for now these prefix are ignored
	 */
	if (((memcmp (dest, &zero6, 16) != 0 || masklen > 0) &&
	     ipv6_compat_addr ((struct in6_addr *) dest)) ||
	    ipv6_link_local_addr ((struct in6_addr *) dest) ||
	    ipv6_multicast_addr ((struct in6_addr *) dest))
	    return (0);
	len = 128;
    }
#endif /* HAVE_IPV6 */

    nexthop = New_Prefix (family, nhop, len);
    prefix = New_Prefix (family, dest, masklen);
    trace (TR_INFO, MRT->trace, "KERNEL %s %s index %d\n",
	   prefix_toax (prefix), prefix_toa (nexthop), index);
    add_route_to_rib (PROTO_KERNEL, prefix, nexthop, interface, KERNEL_PREF, 0);
    Deref_Prefix (prefix);
    Deref_Prefix (nexthop);
    return (1);
}

